# -*- coding: utf-8 -*-

from pulp.server import exceptions
from pulp.server.controllers import importer as importer_controller
from pulp.server.controllers import repository as repo_controller
from pulp.server.db import model
from pulp.server.db.model.dispatch import ScheduledCall
from pulp.server.managers.schedule import utils

_PUBLISH_OPTION_KEYS = ('override_config',)
_SYNC_OPTION_KEYS = ('override_config',)


class RepoSyncScheduleManager(object):
    @classmethod
    def list(cls, repo_id, importer_id):
        """
        Returns an iterator of ScheduledCall instances that represent schedules
        for the specified repo and importer.

        :param repo_id:     unique ID for a repository
        :type  repo_id:     basestring
        :param importer_id: unique ID for an importer
        :type  importer_id: basestring

        :return:    iterator of ScheduledCall instances
        :rtype:     iterator
        """
        importer_controller.get_valid_importer(repo_id, importer_id)
        return utils.get_by_resource(importer_controller.build_resource_tag(repo_id, importer_id))

    @classmethod
    def create(cls, repo_id, importer_id, sync_options, schedule,
               failure_threshold=None, enabled=True):
        """
        Create a new sync schedule for a given repository using the given importer.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param importer_id:     unique ID for an importer
        :type  importer_id:     basestring
        :param sync_options:    dictionary that contains the key 'override_config',
                                whose value should be passed as the 'overrides'
                                parameter to the sync task. This wasn't originally
                                documented, so it isn't clear why overrides value
                                couldn't be passed directly.
        :type  sync_options:    dict
        :param schedule_data:   dictionary that contains the key 'schedule', whose
                                value is an ISO8601 string. This wasn't originally
                                documented, so it isn't clear why the string itself
                                couldn't have been passed directly.
        :type  schedule_data:   dict

        :return:    new schedule instance
        :rtype:     pulp.server.db.model.dispatch.ScheduledCall
        """
        # validate the input
        importer_controller.get_valid_importer(repo_id, importer_id)
        utils.validate_keys(sync_options, _SYNC_OPTION_KEYS)
        utils.validate_initial_schedule_options(schedule, failure_threshold, enabled)

        task = repo_controller.queue_sync_with_auto_publish.name
        args = [repo_id]
        kwargs = {'overrides': sync_options['override_config']}
        resource = importer_controller.build_resource_tag(repo_id, importer_id)
        schedule = ScheduledCall(schedule, task, args=args, kwargs=kwargs,
                                 resource=resource, failure_threshold=failure_threshold,
                                 enabled=enabled)
        schedule.save()
        try:
            importer_controller.get_valid_importer(repo_id, importer_id)
        except exceptions.MissingResource:
            # back out of this whole thing, since the importer disappeared
            utils.delete(schedule.id)
            raise

        return schedule

    @classmethod
    def update(cls, repo_id, importer_id, schedule_id, updates):
        """
        Update an existing sync schedule.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param importer_id:     unique ID for an importer
        :type  importer_id:     basestring
        :param schedule_id:     unique ID for a schedule
        :type  schedule_id:     basestring
        :param updates:         dictionary of updates to apply
        :type  updates:         dict

        :return ScheduledCall instance as it appears after the update
        :rtype  pulp.server.db.model.dispatch.ScheduledCall
        """
        importer_controller.get_valid_importer(repo_id, importer_id)

        # legacy logic that can't be explained
        if 'override_config' in updates:
            updates['kwargs'] = {'overrides': updates.pop('override_config')}

        utils.validate_updated_schedule_options(updates)

        return utils.update(schedule_id, updates)

    @classmethod
    def delete(cls, repo_id, importer_id, schedule_id):
        """
        Delete a scheduled sync from a given repository and importer.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param importer_id:     unique ID for an importer
        :type  importer_id:     basestring
        :param schedule_id:     unique ID for a schedule
        :type  schedule_id:     basestring
        """
        # validate the input
        importer_controller.get_valid_importer(repo_id, importer_id)

        # remove from the scheduler
        utils.delete(schedule_id)

    @staticmethod
    def delete_by_importer_id(repo_id, importer_id):
        """
        Delete all schedules for the specified repo and importer.

        :param importer_id:     unique ID for an importer
        :type  importer_id:     basestring
        """
        utils.delete_by_resource(importer_controller.build_resource_tag(repo_id, importer_id))


class RepoPublishScheduleManager(object):
    @classmethod
    def list(cls, repo_id, distributor_id):
        """
        Returns an iterator of ScheduledCall instances that represent schedules
        for the specified repo and distributor.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param distributor_id:  unique ID for a distributor
        :type  distributor_id:  basestring

        :return:    iterator of ScheduledCall instances
        :rtype:     iterator
        """
        dist = model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        return utils.get_by_resource(dist.resource_tag)

    @classmethod
    def create(cls, repo_id, distributor_id, publish_options, schedule,
               failure_threshold=None, enabled=True):
        """
        Create a new scheduled publish for the given repository and distributor.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param distributor_id:  unique ID for a distributor
        :type  distributor_id:  basestring
        :param publish_options: dictionary that contains the key 'override_config',
                                whose value should be passed as the 'overrides'
                                parameter to the publish task. This wasn't originally
                                documented, so it isn't clear why overrides value
                                couldn't be passed directly.
        :type  sync_options:    dict
        :param schedule_data:   dictionary that contains the key 'schedule', whose
                                value is an ISO8601 string. This wasn't originally
                                documented, so it isn't clear why the string itself
                                couldn't have been passed directly.
        :type  schedule_data:   dict

        :return:    new schedule instance
        :rtype:     pulp.server.db.model.dispatch.ScheduledCall
        """
        dist = model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        utils.validate_keys(publish_options, _PUBLISH_OPTION_KEYS)
        utils.validate_initial_schedule_options(schedule, failure_threshold, enabled)

        task = repo_controller.queue_publish.name
        args = [repo_id, distributor_id]
        kwargs = {'overrides': publish_options['override_config']}
        schedule = ScheduledCall(schedule, task, args=args, kwargs=kwargs,
                                 resource=dist.resource_tag, failure_threshold=failure_threshold,
                                 enabled=enabled)
        schedule.save()

        try:
            model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        except exceptions.MissingResource:
            # back out of this whole thing, since the distributor disappeared
            utils.delete(schedule.id)
            raise

        return schedule

    @classmethod
    def update(cls, repo_id, distributor_id, schedule_id, updates):
        """
        Update an existing scheduled publish for the given repository and distributor.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param distributor_id:  unique ID for a distributor
        :type  distributor_id:  basestring
        :param schedule_id:     unique ID for a schedule
        :type  schedule_id:     basestring
        :param updates:         dictionary of updates to apply
        :type  updates:         dict

        :return ScheduledCall instance as it appears after the update
        :rtype  pulp.server.db.model.dispatch.ScheduledCall
        """
        model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        if 'override_config' in updates:
            updates['kwargs'] = {'overrides': updates.pop('override_config')}

        utils.validate_updated_schedule_options(updates)

        return utils.update(schedule_id, updates)

    @classmethod
    def delete(cls, repo_id, distributor_id, schedule_id):
        """
        Delete an existing scheduled publish from the given repository and distributor.

        :param repo_id:         unique ID for a repository
        :type  repo_id:         basestring
        :param distributor_id:  unique ID for a distributor
        :type  distributor_id:  basestring
        :param schedule_id:     unique ID for a schedule
        :type  schedule_id:     basestring
        """
        model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        utils.delete(schedule_id)

    @staticmethod
    def delete_by_distributor_id(repo_id, distributor_id):
        """
        Delete all schedules for the specified repo and distributor.

        :param repo_id:  unique ID for a repository
        :type  repo_id:  basestring
        :param distributor_id:  unique ID for a distributor
        :type  distributor_id:  basestring
        """
        dist = model.Distributor.objects.get_or_404(repo_id=repo_id, distributor_id=distributor_id)
        utils.delete_by_resource(dist.resource_tag)
